/*
  CatWeasel MK3 PCI Linux driver (c) 2002..2009 by Dirk Jagdmann <doj@cubic.org>

  This file is part of CatWeasel MK3 PCI Linux driver.

  The CatWeasel MK3 PCI Linux driver is free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  The CatWeasel MK3 PCI Linux driver is distributed in the hope that
  it will be useful, but WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
  PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with the CatWeasel MK3 PCI Linux driver; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
  MA 02111-1307 USA

  $Header: /code/cw/device/cwjoy.c,v 1.9.2.2 2009/02/09 22:17:39 doj Exp $
*/

/*
  Code for accessing the joysticks and paddles
*/

#include "cw.h"

#include <asm/io.h>
#include <linux/delay.h>
#include <linux/input.h>
#include <linux/timer.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,26)
typedef unsigned* uintptr_t;
#endif

static int cwjoy_rate=25;
module_param(cwjoy_rate, int, S_IRUGO);
MODULE_PARM_DESC(cwjoy_rate,"update joysticks and paddles x times per second (default=25)");

#define LINUX_INPUT_HAS_BTN_BUG 1

enum {
  Joystick1 = 0,
  Joystick2,
  Paddle1,
  Paddle2,
  Nummer,
};

static struct input_dev* cwjoy_dev[CWSID_MAX_NUM][Nummer];

static struct timer_list cwjoy_timer;

static volatile int run;
enum {
  RUNNING,
  STOPPING,
  STOPPED,
};

void cwjoy_poll(unsigned long l)
{
  int i;

  /* reschedule ourself */
  if(run==RUNNING)
    {
      mod_timer(&cwjoy_timer, jiffies+HZ/cwjoy_rate);
    }
  else if(run==STOPPING)
    {
      /* acknoledge, that we have seen the run/stop flag */
      run=STOPPED;
      return;
    }
  else
    {
      return;
    }

  amikbd_poll();

  for(i=0; i<cw_num; i++)
    {
      static int toggle=0;

      /* joysticks */
      const unsigned char d=inb(cw_register[i][JoyDat]);
      const unsigned char b=inb(cw_register[i][JoyButton]);

      if((uintptr_t)(input_get_drvdata(cwjoy_dev[i][Joystick1])))
	{
	  input_report_abs(cwjoy_dev[i][Joystick1], ABS_X, (d&1?0:1) - (d&2?0:1));
	  input_report_abs(cwjoy_dev[i][Joystick1], ABS_Y, (d&4?0:1) - (d&8?0:1));
	  input_report_key(cwjoy_dev[i][Joystick1], BTN_TRIGGER, !(b&64));
	  input_sync(cwjoy_dev[i][Joystick1]);
	}

      if((uintptr_t)(input_get_drvdata(cwjoy_dev[i][Joystick2])))
	{
	  input_report_abs(cwjoy_dev[i][Joystick2], ABS_X, (d&16?0:1) - (d&32?0:1));
	  input_report_abs(cwjoy_dev[i][Joystick2], ABS_Y, (d&64?0:1) - (d&128?0:1));
	  input_report_key(cwjoy_dev[i][Joystick2], BTN_TRIGGER, !(b&128));
	  input_sync(cwjoy_dev[i][Joystick2]);
	}

      /* paddles */
      if((toggle=!toggle))
	{
	  if((uintptr_t)(input_get_drvdata(cwjoy_dev[i][Paddle1])))
	    {
	      input_report_abs(cwjoy_dev[i][Paddle1], ABS_X, sid_read(0x19, i));
	      input_report_abs(cwjoy_dev[i][Paddle1], ABS_Y, sid_read(0x1A, i));
	      input_sync(cwjoy_dev[i][Paddle1]);
	    }
	  outb(0, cw_register[i][PaddleSelect]); /* switch to paddle 2 */
	}
      else
	{
	  if((uintptr_t)(input_get_drvdata(cwjoy_dev[i][Paddle2])))
	    {
	      input_report_abs(cwjoy_dev[i][Paddle2], ABS_X, sid_read(0x19, i));
	      input_report_abs(cwjoy_dev[i][Paddle2], ABS_Y, sid_read(0x1A, i));
	      input_sync(cwjoy_dev[i][Paddle2]);
	    }
	  inb(cw_register[i][PaddleSelect]); /* switch to paddle 1 */
	}
    }
}

static int open(struct input_dev *dev)
{
  input_set_drvdata(dev, (void*)((uintptr_t)input_get_drvdata(dev) + 1));
  return 0;
}

static void close(struct input_dev *dev)
{
  int ref = (int)input_get_drvdata(dev);
  if(--ref < 0)
    ref = 0;
  input_set_drvdata(dev, (void*)(uintptr_t)ref);
}

static int initialized=0;

int __init cwjoy_init(void)
{
  int i;

  static char str[CWSID_MAX_NUM][4][32];

  for(i=0; i<cw_num; i++)
    {
      cwjoy_dev[i][Joystick1] = input_allocate_device();
      input_set_drvdata(cwjoy_dev[i][Joystick1], 0);
      cwjoy_dev[i][Joystick1]->open=open;
      cwjoy_dev[i][Joystick1]->close=close;
      sprintf(str[i][Joystick1], "CatWeasel %i Joystick 1", i);
      cwjoy_dev[i][Joystick1]->name = str[i][Joystick1];
      cwjoy_dev[i][Joystick1]->phys = str[i][Joystick1];
      cwjoy_dev[i][Joystick1]->uniq = "uniq Joystick 1";
      cwjoy_dev[i][Joystick1]->id.bustype = BUS_PCI;
      cwjoy_dev[i][Joystick1]->id.vendor = 0xE159;
      cwjoy_dev[i][Joystick1]->id.product = 1;
      cwjoy_dev[i][Joystick1]->id.version = 0x0100;
      cwjoy_dev[i][Joystick1]->evbit[0] = BIT(EV_KEY) | BIT(EV_ABS);
      set_bit(BTN_TRIGGER, cwjoy_dev[i][Joystick1]->keybit);
      set_bit(ABS_X, cwjoy_dev[i][Joystick1]->absbit);
      cwjoy_dev[i][Joystick1]->absmin[ABS_X] = -1;
      cwjoy_dev[i][Joystick1]->absmax[ABS_X] = +1;
      set_bit(ABS_Y, cwjoy_dev[i][Joystick1]->absbit);
      cwjoy_dev[i][Joystick1]->absmin[ABS_Y] = -1;
      cwjoy_dev[i][Joystick1]->absmax[ABS_Y] = +1;
      input_register_device(cwjoy_dev[i][Joystick1]);

      cwjoy_dev[i][Joystick2] = input_allocate_device();
      input_set_drvdata(cwjoy_dev[i][Joystick2], 0);
      cwjoy_dev[i][Joystick2]->open=open;
      cwjoy_dev[i][Joystick2]->close=close;
      sprintf(str[i][Joystick2], "CatWeasel %i Joystick 2", i);
      cwjoy_dev[i][Joystick2]->name = str[i][Joystick2];
      cwjoy_dev[i][Joystick2]->id.bustype = BUS_PCI;
      cwjoy_dev[i][Joystick2]->id.vendor = 0xE159;
      cwjoy_dev[i][Joystick2]->id.product = 1;
      cwjoy_dev[i][Joystick2]->id.version = 0x0100;
      cwjoy_dev[i][Joystick2]->evbit[0] = BIT(EV_KEY) | BIT(EV_ABS);
      cwjoy_dev[i][Joystick2]->absbit[0] = BIT(ABS_X) | BIT(ABS_Y);
      set_bit(BTN_TRIGGER, cwjoy_dev[i][Joystick2]->keybit);
      cwjoy_dev[i][Joystick2]->absmin[ABS_X] = -1; cwjoy_dev[i][Joystick2]->absmax[ABS_X] = 1;
      cwjoy_dev[i][Joystick2]->absmin[ABS_Y] = -1; cwjoy_dev[i][Joystick2]->absmax[ABS_Y] = 1;
      input_register_device(cwjoy_dev[i][Joystick2]);

      cwjoy_dev[i][Paddle1] = input_allocate_device();
      input_set_drvdata(cwjoy_dev[i][Paddle1], 0);
      cwjoy_dev[i][Paddle1]->open=open;
      cwjoy_dev[i][Paddle1]->close=close;
      sprintf(str[i][Paddle1], "CatWeasel %i Paddle 1", i);
      cwjoy_dev[i][Paddle1]->name = str[i][Paddle1];
      cwjoy_dev[i][Paddle1]->id.bustype = BUS_PCI;
      cwjoy_dev[i][Paddle1]->id.vendor = 0xE159;
      cwjoy_dev[i][Paddle1]->id.product = 1;
      cwjoy_dev[i][Paddle1]->id.version = 0x0100;
#if LINUX_INPUT_HAS_BTN_BUG
      cwjoy_dev[i][Paddle1]->evbit[0] = BIT(EV_ABS) | BIT(EV_KEY);
      set_bit(BTN_TRIGGER, cwjoy_dev[i][Paddle1]->keybit);
#else
      cwjoy_dev[i][Paddle1]->evbit[0] = BIT(EV_ABS);
#endif
      cwjoy_dev[i][Paddle1]->absbit[0] = BIT(ABS_X) | BIT(ABS_Y);
      cwjoy_dev[i][Paddle1]->absmin[ABS_X] = 0; cwjoy_dev[i][Paddle1]->absmax[ABS_X] = 0xFF;
      cwjoy_dev[i][Paddle1]->absmin[ABS_Y] = 0; cwjoy_dev[i][Paddle1]->absmax[ABS_Y] = 0xFF;
      input_register_device(cwjoy_dev[i][Paddle1]);

      cwjoy_dev[i][Paddle2] = input_allocate_device();
      input_set_drvdata(cwjoy_dev[i][Paddle2], 0);
      cwjoy_dev[i][Paddle2]->open=open;
      cwjoy_dev[i][Paddle2]->close=close;
      sprintf(str[i][Paddle2], "CatWeasel %i Paddle 2", i);
      cwjoy_dev[i][Paddle2]->name = str[i][Paddle2];
      cwjoy_dev[i][Paddle2]->id.bustype = BUS_PCI;
      cwjoy_dev[i][Paddle2]->id.vendor = 0xE159;
      cwjoy_dev[i][Paddle2]->id.product = 1;
      cwjoy_dev[i][Paddle2]->id.version = 0x0100;
#if LINUX_INPUT_HAS_BTN_BUG
      cwjoy_dev[i][Paddle2]->evbit[0] = BIT(EV_ABS) | BIT(EV_KEY);
      set_bit(BTN_TRIGGER, cwjoy_dev[i][Paddle2]->keybit);
#else
      cwjoy_dev[i][Paddle2]->evbit[0] = BIT(EV_ABS);
#endif
      cwjoy_dev[i][Paddle2]->absbit[0] = BIT(ABS_X) | BIT(ABS_Y);
      cwjoy_dev[i][Paddle2]->absmin[ABS_X] = 0; cwjoy_dev[i][Paddle2]->absmax[ABS_X] = 0xFF;
      cwjoy_dev[i][Paddle2]->absmin[ABS_Y] = 0; cwjoy_dev[i][Paddle2]->absmax[ABS_Y] = 0xFF;
      input_register_device(cwjoy_dev[i][Paddle2]);
    }

  /* enable timer */
  run=RUNNING;
  init_timer(&cwjoy_timer);
  cwjoy_timer.function=cwjoy_poll;
  cwjoy_timer.expires=jiffies+HZ/cwjoy_rate;
  add_timer(&cwjoy_timer);

  initialized=1;
  return 0;
}

void __exit cwjoy_exit(void)
{
  if(initialized)
    {
      int c;

      del_timer_sync(&cwjoy_timer);

      for(c=cw_num-1; c>=0; c--)
	{
	  int i;
	  for(i=Nummer-1; i>=0; i--)
	    {
	      input_unregister_device(cwjoy_dev[c][i]);
	    }
	}

      initialized=0;
    }
}
