Program Catweaseltester;
uses crt;
const base=$ec00;
(*
make that a PnP recognition routine..!
Catweasel MK4 has vendor ID $e159, subsystem vendor ID $5213
(notice the difference to the MK3 - choice of an unassigned number)
simple communications controller (just like the MK3)
*)

var f: file of byte;

(*
base+2 is data direction
base+3 is data port output direction
base+7 is data port input direction

bit7: highbit of register bank select (set to out)
bit6: lowbit of register bank select (set to out)
bit5: multiplex (set to out)
bit4: joystick 2 firebutton (can be in or out)
bit3: IRQ (set to in - in config mode: Ready indicator)
bit2: Config_done (set to in, 0 indicates empty FPGA)
bit1: Legacy (set to out, 1 sets control/status registers bit order to
             MK1, 0 sets the bit orders of those two registers to MK2
             or higher. Legacy also plays a role in FPGA configuration)
bit0: nConfig (set to out, 0 will clear the chip, 1 will let it run)

When FPGA is unconfigured, only register bank 01 can be used.
After configuring, all register banks are defined by the core itself,
see core documentation.
*)

procedure InitPCIbridge;
begin
     port[base]:=241;
     port[base+1]:=0;
     port[base+2]:=227; (* see above, 0=in, 1=out *)
     port[base+3]:=65; (* dont reset, select config data bank *)
     port[base+4]:=0;
     port[base+5]:=0;
     port[base+$29]:=0;
     port[base+$2b]:=0;
end;

Function FPGAready:Boolean;
begin
  if (port[base+7] AND 8)=8 then FPGAready:=true else FPGAready:=false;
end;

Function Config_done:Boolean;
begin
  if (port[base+7] AND 4)=4 then Config_done:=true else Config_done:=false;
end;

Procedure ResetFPGA;
Begin
  port[base+2]:=227;
  port[base+3]:=0; (* reset FPGA *)
  delay(1); (* 1ms delay *)
  port[base+3]:=65; (* get it out of reset state and select register bank *)

end;

Procedure OpenFile;
Begin
  assign (f,'c:\cwmk4\corev001.cw4');
  reset (f);
end;

Procedure Closefile;
Begin
  close (f);
end;

Procedure WriteConfigData;
var b: byte;
    i: longint;
Begin
  Openfile;

  for i:=1 to 59215 do
    begin
      read(f,b);
      if (b AND 1)=1 then port[base+3]:=67 else port[base+3]:=65;
      repeat until FPGAready; (* wait for FPGA to accept data *)
      port[base+192]:=b;
    end;

  Closefile;
end;

Procedure checkmem;
var i,dummy:byte;
Begin
dummy:=port[base+196]; (*reset pointer *)

port[base+192]:=0; (* write a few bytes *)
port[base+192]:=1;
port[base+192]:=2;
port[base+192]:=4;
port[base+192]:=8;
port[base+192]:=16;
port[base+192]:=32;
port[base+192]:=64;
port[base+192]:=128;
port[base+192]:=44;
port[base+192]:=22;
port[base+192]:=0;

dummy:=port[base+196]; (* reset the pointer again *)

for i:=1 to 12 do writeln (port[base+192]); (* and view the bytes *)

end;

(* main program starts here *)
begin

InitPCIbridge;
if Config_done then writeln ('entry state: FPGA configured.') else writeln (' entry state: FPGA empty.');
ResetFPGA;
if Config_Done then writeln('PANIC: FPGA did not reset!') else writeln('FPGA softreset OK');
WriteConfigData;
if Config_Done then writeln('FPGA configured OK') else writeln('error configuring FPGA');
checkmem;

end.
