/*
  sid (c) 1999-2004 by Dirk Jagdmann <doj@cubic.org>

  This file is part of sid.

  sid is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  sid is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with sid; if not, write to the Free Software Foundation, Inc.,
  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
  
  $Header: /code/cw/sid/debugger.cpp,v 1.6 2004/05/14 14:51:03 doj Exp $
*/

#include <signal.h>
#include <stdlib.h>

#include <iostream>
#include <sstream>
#include <string>

using namespace std;

#include "c64.h"
#include "stilinfo.hpp"

namespace {

const char CBMchar[128] = {
  /* 00 */ '@','A','B','C','D','E','F','G',
  /* 08 */ 'H','I','J','K','L','M','N','O',
  /* 10 */ 'P','Q','R','S','T','U','V','W',
  /* 18 */ 'X','Y','Z','[','',']','^','.',
  /* 20 */ ' ','!','"','#','$','%','&','\'',
  /* 28 */ '(',')','*','+',',','-','.','/',
  /* 30 */ '0','1','2','3','4','5','6','7',
  /* 38 */ '8','9',':',';','<','=','>','?',
  /* 40 */ 'x','x','x','x','x','x','x','x',
  /* 48 */ 'x','x','x','x','x','x','x','x',
  /* 50 */ 'x','x','x','x','x','x','x','x',
  /* 58 */ 'x','x','x','x','x','x','x','x',
  /* 60 */ 'x','x','x','x','x','x','x','x',
  /* 68 */ 'x','x','x','x','x','x','x','x',
  /* 70 */ 'x','x','x','x','x','x','x','x',
  /* 78 */ 'x','x','x','x','x','x','x','x',
};

const char PETchar[128] = {
  /* 00 */ '@','a','b','c','d','e','f','g',
  /* 08 */ 'h','i','j','k','l','m','n','o',
  /* 10 */ 'p','q','r','s','t','u','v','w',
  /* 18 */ 'x','y','z','[','',']','^','.',
  /* 20 */ ' ','!','"','#','$','%','&','\'',
  /* 28 */ '(',')','*','+',',','-','.','/',
  /* 30 */ '0','1','2','3','4','5','6','7',
  /* 38 */ '8','9',':',';','<','=','>','?',
  /* 40 */ '.','A','B','C','D','E','F','G',
  /* 48 */ 'H','I','J','K','L','M','N','O',
  /* 50 */ 'P','Q','R','S','T','U','V','W',
  /* 58 */ 'X','Y','Z','.','.','.','.','.',
  /* 60 */ ' ','.','.','.','.','.','.','.',
  /* 68 */ '.','.','.','.','.','.','.','.',
  /* 70 */ '.','.','.','.','.','.','.','.',
  /* 78 */ '.','.','.','.','.','.','.','.',
};

void pcaxy(istream& i, int& pc, int& a, int& x, int& y)
{
  string s;
  getline(i, s);
  istringstream is(s);

  is >> s;
  if(is)
    {
      pc=strtol(s.c_str(), 0, 16);
      is >> s;
      if(is)
	{
	  a=strtol(s.c_str(), 0, 16);
	  is >> s;
	  if(is)
	    {
	      x=strtol(s.c_str(), 0, 16);
	      is >> s;
	      if(is)
		{
		  y=strtol(s.c_str(), 0, 16);
		}
	    }
	}
    }
}

  MOS6510 *CPU;

void stopcpu(int s)
{
  if(CPU)
    CPU->abort=true;
}

};

int C64::debug(const word npc, const byte na, const byte nx, const byte ny)
{
  int ret=0;

  cpu_.set(npc, na, nx, ny);
  cpu_.debug=true;
  cpu_.info();
  CPU=&cpu_;

  sighandler_t int_h=signal(SIGINT, stopcpu);
  sighandler_t term_h=signal(SIGTERM, stopcpu);

  string oldcommand;
  while(true)
    {
      string line; getline(cin, line);
      istringstream is(line);
      string command;
      if(line.size())
	{
	  is >> command;
	  oldcommand=command;
	}
      else
	command=oldcommand;

      if(command=="q" || command=="x")
	break;
      else if(command=="Q")
	{
	  ret=-1;
	  break;
	}
      // assemble
      else if(command=="a")
	{
	  cerr << "a not implemented" << endl;
	}
      // toggle breakpoint
      else if(command=="b")
	{
	  int pc=-1, a=-1, x=-1, y=-1;
	  pcaxy(is, pc, a, x, y);
	  cpu_.breakpointToggle(pc);
	  cpu_.info();
	}
      // print breakpoints
      else if(command=="bp")
	{
	  cpu_.breakpointPrint();
	  cpu_.info();
	}
      // clear breakpoints
      else if(command=="bc")
	{
	  cpu_.breakpointClear();
	  cpu_.info();
	}
      // compare
      else if(command=="c")
	{
	  cerr << "c not implemented" << endl;
	}
      // disassemble
      else if(command=="d")
	{
	  static int pcs=-1;
	  int pc=-1, a=-1, x=-1, y=-1;
	  pcaxy(is, pc, a, x, y);
	  if(pc==-1)
	    pc=pcs;
	  pcs=cpu_.disass(pc, stdout, false);
	  putchar('\t');
	  fflush(stdout);
	}
      // fill
      else if(command=="f")
	{
	  int pc=-1, a=-1, x=-1, y=-1;
	  pcaxy(is, pc, a, x, y);
	  if(x==-1)
	    x=1;
	  if(pc<0 || pc>0xFFFF)
	    {
	      printf("addr out of range: %04X\n", pc);
	      goto nofill;
	    }
	  if(pc+x-1>0xFFFF)
	    {
	      printf("size exceeds 0xFFFF\n");
	      goto nofill;
	    }
	  if(a<0 || a>0xFF)
	    {
	      printf("value of of range: %X\n", a);
	      goto nofill;
	    }
	  memset(memory()+pc, a, x);
	  printf("filled %X bytes\n", x);
	nofill: ;
	}
      // go
      else if(command=="g")
	{
	  int pc=-1, a=-1, x=-1, y=-1;
	  pcaxy(is, pc, a, x, y);
	  infostr.clear();
	  int cyc=cpu_.jsr(pc, a, x, y);
	  cout << "executed " << cyc << " cycles" << endl;
	  for(infostr_t::iterator i=infostr.begin(); i!=infostr.end(); ++i)
	    cout << *i << endl;

	  cpu_.info();
	}
      // single step
      else if(command==".")
	{
	  int pc=-1, a=0, x=0, y=0;
	  pcaxy(is, pc, a, x, y);
	  if(pc>=0 && pc<=0xFFFF)
	    cpu_.set(pc,a,x,y);
	  infostr.clear();
	  int cyc=cpu_.parse();
	  cout << "executed " << cyc << " cycles" << endl;
	  for(infostr_t::iterator i=infostr.begin(); i!=infostr.end(); ++i)
	    cout << *i << endl;
	  cpu_.info();
	}
      else if(command=="pc")
	{
	  int pc=-1, a=0, x=0, y=0;
	  pcaxy(is, pc, a, x, y);
	  if(pc>=0 && pc<=0xFFFF)
	    cpu_.set(pc,a,x,y);
	  cpu_.info();
	}
      // hunt
      else if(command=="h")
	{
	  cerr << "h not implemented" << endl;
	}
      // load
      else if(command=="l")
	{
	  int addr;
	  FILE *f;
	  size_t size;
	  string fn, addrs;
	  is >> fn;
	  if(fn.empty())
	    {
	      cerr << "no filename" << endl;
	      goto noload;
	    }
	  is >> addrs;
	  addr=strtol(addrs.c_str(), 0, 16);
	  if(addr<0 || addr>0xFFFF)
	    {
	      cerr << "load address out of range: " << addr << endl;
	      goto noload;
	    }

	  f=fopen(fn.c_str(), "rb");
	  if(f==0)
	    {
	      perror("could not open file");
	      goto noload;
	    }

	  if(fseek(f, 0, SEEK_END) < 0)
	    {
	      perror("could not seek");
	      goto abortload;
	    }

	  size=ftell(f);
	  rewind(f);

	  if(addr+size-1>0xFFFF)
	    {
	      cerr << "size of file would exceed 0xFFFF: " << size << endl;
	      goto abortload;
	    }

	  if(fread(memory()+addr, size, 1, f)!=1)
	    {
	      perror("could not read entire file");
	    }

	  printf("loaded $%04X bytes\n", size);

	abortload:
	  fclose(f);
	noload: ;
	}
      // memory display
      else if(command=="m")
	{
	  int pc=-1, a=-1, x=0, y=0;
	  pcaxy(is, pc, a, x, y);
	  if(a==-1)
	    a=0x10;
	  if(pc>=0 && pc<=0xFFFF)
	    {
	      printf("%04X: ", pc);
	      for(; pc<=0xFFFF && a; pc++, a--)
		printf("%02X ", getmem(pc));
	      printf("\n");
	    }
	}
      // registers
      else if(command=="r")
	{
	  cpu_.info();
	}
      // save
      else if(command=="s")
	{
	  cerr << "s not implemented" << endl;
	}
      // copy/transport
      else if(command=="t")
	{
	  cerr << "t not implemented" << endl;
	}
      // toggle memory watch
      else if(command=="w")
	{
	  int pc=-1, a=-1, x=0, y=0;
	  pcaxy(is, pc, a, x, y);
	  if(pc<0 || pc>0xFFFF)
	    printf("invalid address %x\n", pc);
	  else
	    watchToggle(pc);
	}
      // clear watchpoints
      else if(command=="wc")
	{
	  watchClear();
	  printf("cleared watchpoints\n");
	}
      // print watchpoints
      else if(command=="wp")
	{
	  watchPrint();
	}
      // help
      else if(command=="?")
	{
	  cout << "a - assemble\n";
	  cout << "b addr - toggle breakpoint\n";
	  cout << "bp - print breakpoints\n";
	  cout << "bc - clear all breakpoints\n";
	  cout << "c - compare\n";
	  cout << "d [addr] - disasseble\n";
	  cout << "f addr byte size - fill memory\n";
	  cout << "g addr a x y - go\n";
	  cout << ". addr a x y - single step\n";
	  cout << "h - hunt\n";
	  cout << "l filename - load\n";
	  cout << "m addr size - display memory\n";
	  cout << "pc addr a x y - set registers to\n";
	  cout << "r - print registers\n";
	  cout << "s - save\n";
	  cout << "t - transport\n";
	  cout << "w addr - memory watch\n";
	  cout << "wc - clear memory watchpoints\n";
	  cout << "wp - print memory watchpoints\n";
	  cout << "dex2hex num - convert num\n";
	  cout << "hex2dec num - convert num\n";
	  cout << "reset - reset cpu\n";
	  cout << "screen ['p'] - print textmode screen, parameter 'p' for PET charmap\n";

	  cpu_.info();
	}
      // reset cpu
      else if(command=="reset")
	{
	  cpu_.reset();
	  cpu_.info();
	}
      else if(command=="dec2hex")
	{
	  string s; is >> s;
	  int a=atoi(s.c_str());
	  printf("%X\n", a);
	}
      else if(command=="hex2dec")
	{
	  string s; is >> s;
	  int a=strtol(s.c_str(), 0, 16);
	  printf("%i\n", a);
	}
      // print current text screen
      else if(command=="screen")
	{
	  string s;
	  is >> s;
	  const char *t=(s=="p")?PETchar:CBMchar;

	  const int bank=(getmem(0xDD00)^3)&3;
	  const int base=getmem(0xD018)>>4;
	  const int offset=bank*0x4000+base*0x0400;
	  printf("screen at %04X\n", offset);
	  byte *p=memory()+offset;
	  for(int y=0; y<25; y++)
	    {
	      for(int x=0; x<40; x++)
		printf("%c", t[(*p++)&0x7F]);
	      printf("\n");
	    }
	}
    }

  signal(SIGINT, int_h);
  signal(SIGTERM, term_h);
  CPU=0;
  cpu_.debug=cpu_.abort=false;

  return ret;
}
